import sys

sys.path.append(".")
sys.path.append("..")

import os
import json
import time
import xml.etree.ElementTree as ET
import logging

logging.basicConfig(level=logging.DEBUG)

import numpy as np
from shapely.geometry import Point
from PIL import Image
import pygame
import pytest

from tactics2d.map.parser import Lanelet2Parser
from tactics2d.trajectory.parser import LevelXParser
from tactics2d.scenario.render_manager import RenderManager
from tactics2d.sensor import TopDownCamera, SingleLineLidar


@pytest.mark.render
@pytest.mark.parametrize("follow_view", [True, False])
def test_camera(follow_view: bool):
    map_path = "./tactics2d/data/map_default/I_0_inD_DEU.osm"
    trajectory_path = "./tactics2d/data/trajectory_sample/inD/data/"
    config_path = "./tactics2d/data/map_default.config"

    with open(config_path, "r") as f:
        configs = json.load(f)

    map_parser = Lanelet2Parser()
    map_root = ET.parse(map_path).getroot()
    map_ = map_parser.parse(map_root, configs["I_0"])

    frame = 40
    trajectory_parser = LevelXParser("inD")
    participants = trajectory_parser.parse(0, trajectory_path, (0.0, 200.0))
    participant_ids = [
        participant.id_
        for participant in participants.values()
        if participant.is_active(frame)
    ]

    if follow_view:
        camera = TopDownCamera(1, map_, window_size=(600, 600))
        camera.update(participants, participant_ids, frame)
    else:
        camera = TopDownCamera(1, map_, (30, 30, 45, 15), window_size=(600, 600))
        state = participants[participant_ids[0]].get_state(frame)
        camera.update(
            participants, participant_ids, frame, Point(state.location), state.heading
        )
    observation = camera.get_observation()
    logging.info(f"observation.shape: {observation.shape}")

    img = Image.fromarray(observation)
    img = img.rotate(270)
    if follow_view:
        img.save("./tests/runtime/test_camera_follow_view.jpg")
    else:
        img.save("./tests/runtime/test_camera.jpg")


@pytest.mark.render
@pytest.mark.parametrize("perception_range", [12.0, 30.0, 45.0, 100.0])
def test_lidar(perception_range):
    map_path = "./tactics2d/data/map_default/I_0_inD_DEU.osm"
    trajectory_path = "./tactics2d/data/trajectory_sample/inD/data/"
    config_path = "./tactics2d/data/map_default.config"

    with open(config_path, "r") as f:
        configs = json.load(f)

    map_parser = Lanelet2Parser()
    map_root = ET.parse(map_path).getroot()
    map_ = map_parser.parse(map_root, configs["I_0"])

    frame = 40
    trajectory_parser = LevelXParser("inD")
    participants = trajectory_parser.parse(0, trajectory_path, (0.0, 200.0))
    participant_ids = [
        participant.id_
        for participant in participants.values()
        if participant.is_active(frame)
    ]

    lidar = SingleLineLidar(
        1, map_, perception_range, window_size=(600, 600), off_screen=False
    )

    state = participants[participant_ids[0]].get_state(frame)
    lidar.update(
        participants, participant_ids[1:], frame, Point(state.location), state.heading
    )
    _ = lidar.get_observation()

    observation = pygame.surfarray.array3d(lidar.surface)
    img = Image.fromarray(observation)
    img = img.rotate(270)

    img.save(f"./tests/runtime/test_lidar_{int(perception_range)}.jpg")


@pytest.mark.render
@pytest.mark.skipif("DISPLAY" not in os.environ, reason="requires display server")
@pytest.mark.parametrize(
    "layout_style, off_screen",
    [("block", False), ("hierarchical", False), ("block", True), ("hierarchical", True)],
)
def test_render_manager(layout_style, off_screen):
    """This function tests the following functions in RenderManager:
    _rearrange_layout, add, is_bound, bind, unbind, remove_sensor, update, render, close
    """
    map_path = "./tactics2d/data/map_default/I_0_inD_DEU.osm"
    trajectory_path = "./tactics2d/data/trajectory_sample/inD/data/"
    config_path = "./tactics2d/data/map_default.config"

    with open(config_path, "r") as f:
        configs = json.load(f)

    map_parser = Lanelet2Parser()
    map_root = ET.parse(map_path).getroot()
    map_ = map_parser.parse(map_root, configs["I_0"])

    trajectory_parser = LevelXParser("inD")
    participants = trajectory_parser.parse(0, trajectory_path, (0.0, 200.0))

    render_manager = RenderManager(
        fps=100, windows_size=(600, 600), layout_style=layout_style, off_screen=off_screen
    )

    perception_range = (30, 30, 45, 15)
    main_camera = TopDownCamera(1, map_, window_size=(600, 600), off_screen=off_screen)
    camera1 = TopDownCamera(
        2,
        map_,
        perception_range=perception_range,
        window_size=(200, 200),
        off_screen=off_screen,
    )
    camera2 = TopDownCamera(
        3,
        map_,
        perception_range=perception_range,
        window_size=(200, 200),
        off_screen=off_screen,
    )

    render_manager.add_sensor(main_camera, main_sensor=True)
    render_manager.add_sensor(camera1)
    render_manager.add_sensor(camera2)

    def auto_bind_camera(camera, participant_ids, bind_target):
        bind_id = render_manager.is_bound(camera.id_)
        if bind_id is None:
            render_manager.bind(camera.id_, participant_ids[bind_target])
        elif not participants[bind_id].is_active(frame):
            render_manager.unbind(camera.id_)
            render_manager.bind(camera.id_, participant_ids[bind_target])

    t1 = time.time()

    for frame in np.arange(0, 50 * 1000, 40):
        participant_ids = [
            participant.id_
            for participant in participants.values()
            if participant.is_active(frame)
        ]

        auto_bind_camera(camera1, participant_ids, 0)
        auto_bind_camera(camera2, participant_ids, 1)

        render_manager.update(participants, participant_ids, frame)
        render_manager.render()

    render_manager.close()

    t2 = time.time()

    average_fps = 1 // ((t2 - t1) / (50 * 1000 / 40))

    logging.debug(f"The average FPS is {average_fps}")


@pytest.mark.skip(reason="This test is not implemented yet.")
def test_off_screen_rendering():
    return
